unit untSkypeChatSynch;

interface

uses Classes, forms;

type
  //------- participants
  TParticipants = record
    people: array [0..1] of string;
    id: array [0..1] of string;
  end;
  
  //------- skype chat synch file
  TSkypeChatSyncFile = record
    fileName: string;
    fileDateTimeStamp: integer;
    fileDateTime: TDateTime;
  end;

  //------- all synch files
  TSkypeChatSyncFiles = array of TSkypeChatSyncFile;

  //------- conversation
  TSkypeConversation = record
     participants: TParticipants;
     content: TStringList;
  end;

  //------- all conversations
  TSkypeConversations = array of TSkypeConversation;

function GetSkypeChatSyncFiles(dataDir: string): TSkypeChatSyncFiles;
function getConvParticipants(chatSyncFName: string): TParticipants;
procedure getConversation(chatSyncFName: string; var myOutputList: TStringList; p: TParticipants; ownName: string);
function retrieveConversations(skypeFiles: TSkypeChatSyncFiles; ownID: string): TSkypeConversations;

function ConvertStreamToASCIIHex(const Stream: TStream): string;
function ConvertASCIIHexToString(dataIn: string): string;

implementation

uses SysUtils;







//------------------------------------------converting TMemoryStream to ASCIIHex
function ConvertStreamToASCIIHex(const Stream: TStream): string;
var
  buffer: string;
  i: integer;
begin
  result := '';
  Stream.Position := 0;
  SetLength(buffer, Stream.Size);
  Stream.ReadBuffer(Pointer(buffer)^, Stream.Size);
  for I := 1 to length(buffer) do result := result + inttohex(ord(buffer[i]), 2) + ' ';
end;






//------------------------------------decrypt encrypted and ASCII-encoded String
function ConvertASCIIHexToString(dataIn: string): string;
var
  i: integer;
begin
  result := '';
  for i := 0 to (length(dataIn) div 3) - 1 do result := result + char(strToInt('$' + copy(dataIn, i * 3 + 1, 2)));
end;




function RemoveSpaces(dataIn: string): string;
var
  rf: TReplaceFlags;
begin
  rf := [rfReplaceAll];
  result := stringReplace(dataIn, ' ', '', rf);
end;





//---------------------------------check if extracted text is valid conversation
function conversationIsValid(dataIn: string): boolean;
var  i: integer;
     b: byte; 
begin
  //--- check if the length of the string is OK
  if length(dataIn) > 0 then result := true else
  begin
    result := false;
    exit;
  end;

  //--- check if there are valid characters only (no support for unicode now...)
  for i := 1 to length(dataIn) do
  begin
    b := ord(dataIn[i]);
    if (b < 32) or (b > 127) then
    begin
      result := false;
      exit;
    end; 
  end;

end;






//----------------------------------------get list of all Skype chat synch files
function GetSkypeChatSyncFiles(dataDir: string): TSkypeChatSyncFiles;
var
  dateStamps: array of integer;

  procedure FindFiles(dataDir: string);
  var
    Rec: TSearchRec;
    intFileAge: LongInt;
  begin
    dataDir := IncludeTrailingPathDelimiter(dataDir);

    //--- this is the file
    if FindFirst(dataDir + '*.*', faAnyFile - faDirectory, Rec) = 0 then
    begin
      try
        repeat
            setLength(result, length(result) + 1);
            result[length(result) - 1].fileName := dataDir + Rec.Name;
            intFileAge := FileAge(result[length(result) - 1].fileName);
            result[length(result) - 1].fileDateTimeStamp := intFileAge;
            result[length(result) - 1].fileDateTime := FileDateToDateTime(intFileAge);

            setLength(dateStamps, length(dateStamps) + 1);
            dateStamps[length(dateStamps) - 1] := intFileAge;
        until FindNext(Rec) <> 0;
      finally
        FindClose(Rec) ;
      end;
    end else

    //--- this is the folder
    if FindFirst(dataDir + '*.*', faDirectory, Rec) = 0 then
    begin
      try
        repeat
          if (Rec.Name <> '.') and (Rec.Name <> '..') then  FindFiles(dataDir + Rec.Name);
        until FindNext(Rec) <> 0;
      finally
        FindClose(Rec) ;
      end;
    end;

  end;

  //--- array sorting algorythm
  procedure QuickSort(var A: TSkypeChatSyncFiles; iLo, iHi: Integer);
  var
    Lo, Hi, Mid: Integer;
    TmpFile: TSkypeChatSyncFile;
  begin
    Lo := iLo;
    Hi := iHi;
    Mid := A[(Lo + Hi) div 2].fileDateTimeStamp;
    repeat
      while A[Lo].fileDateTimeStamp < Mid do Inc(Lo);
      while A[Hi].fileDateTimeStamp > Mid do Dec(Hi);
      if Lo <= Hi then
      begin
        TmpFile := A[Lo];
        A[Lo] := A[Hi];
        A[Hi] := TmpFile;
        Inc(Lo);
        Dec(Hi);
      end;
    until Lo > Hi;
    if Hi > iLo then QuickSort(A, iLo, Hi);
    if Lo < iHi then QuickSort(A, Lo, iHi);
  end;

begin
  FindFiles(dataDir);
  QuickSort(result, 0, high(dateStamps));
end;






//----------------read information about participants from Skype chat synch file
function getConvParticipants(chatSyncFName: string): TParticipants;
const
  tokenSearchForID = '00 00 01 ';
var
  fs: TFileStream;
  rf: TReplaceFlags;
  c: byte;
  l: string;
  findPos1: integer;
begin
  if ExtractFileExt(chatSyncFName) = '.dat' then
  begin
      l := '';
      rf := [rfReplaceAll];
      fs := TFileStream.Create(chatSyncFName, fmOpenRead or fmShareDenyNone);
      fs.Seek($35, soFromBeginning);
      repeat
         fs.ReadBuffer(c, 1);
         if c <> $3B then l := l + char(c);
      until c = $3B;
      l := StringReplace(l, '/$', char($09), rf);
      result.people[0] := trim(copy(l, 1, pos(char($09), l)));
      result.people[1] := trim(copy(l, pos(char($09), l) + 1, length(l)));

      //--- get own identifier (ID of the user who is running skype)
      l := ConvertStreamToASCIIHex(fs);
      findPos1 := pos(tokenSearchForID, l);

      result.id[0] := copy(l, findPos1 + 5*3, 6);
      fs.Free;
  end;
end;





//------------------return the correct participant name based on his fingerprint
function MatchParticipantsById(p: TParticipants; Id, ownName: string): string;
begin
  if ownName <> '' then //--- there is own name provided
  begin
    if id = p.id[0] then result := ownName else
      if ownName = p.people[0] then result := p.people[1] else result := p.people[0];
  end else

  begin //--- own name is not provided
    if id = p.id[0] then result := 'user 1' else  result := 'user 2';
  end;
  //result := result + ' ' + Id + '(' + p.id[0] + ')'; //--- needed for tests only
end;




//------------------function returns the offset for certain token, searched back
function backSearch(dataIn: string; initialPos, searchLen: integer; token1, token2, token3: string): integer;
var i: integer;
    mySubstr1,
    mySubstr2,
    mySubstr3: string;
begin
  result := -1;
  for i := 1 to searchLen do
  begin
    mySubstr1 := copy(dataIn, initialPos - i * 3, length(token1)); //--- first token should be: 4x
    mySubstr2 := copy(dataIn, initialPos - i * 3 + 3, length(token2)); //-- second token should be: 01
    mySubstr3 := copy(dataIn, initialPos - i * 3 + 3*7, 9); //--- third token should be: 00 00 00
    if (mySubstr1 = token1) and (mySubstr2 = token2) and (mySubstr3 = token3) then
    begin
      result := initialPos - i * 3;
      break;
    end;
  end;
end;




//------------------------------read one conversation from Skype chat synch file
procedure getConversation(chatSyncFName: string; var myOutputList: TStringList; p: TParticipants; ownName: string);
const
  tokenConvStart = '03 02 ';
  tokenConvEnd = '00 ';
  tokenConvHeaderStart1 = '4';
  tokenConvHeaderStart2 = '01 ';
  tokenConvHeaderStart3 = '00 00 00 ';

  // Sets UnixStartDate to TDateTime of 01/01/1970 
  UnixStartDate: TDateTime = 25569.0; 

var
  fs: TFileStream;
  rf: TReplaceFlags;
  l, messg, timestamp, conversation, currentUserId, selector: string;
  findPos1,
  findPos2,
  timestampOffset: integer;
begin
  if ExtractFileExt(chatSyncFName) = '.dat' then
  begin
      rf := [rfReplaceAll];
      fs := TFileStream.Create(chatSyncFName, fmOpenRead or fmShareDenyNone);
      l := ConvertStreamToASCIIHex(fs);
      fs.Free;

      repeat
         findPos1 := pos(tokenConvStart, l); //--- search for the place where the conversation starts
         if findPos1 > 0 then //--- found! this is where the message starts
         begin
           //--- look for the user's identifier and translate it to the user's name
           selector := copy(l, findPos1 - 3, 2);
           if (selector = '01') or (selector = '02') or (selector = '03')
           or (selector = '04') or (selector = '05') or (selector = '06')
           or (selector = '07') or (selector = '08') or (selector = '09') then //--- this is a dynamic offset, babe...
               currentUserId := copy(l, findPos1 - 9, 6) else
               currentUserId := copy(l, findPos1 - 6, 6);

           //--- now get the message timestamp
           timestampOffset := backSearch(l, findPos1, 255, tokenConvHeaderStart1, tokenConvHeaderStart2, tokenConvHeaderStart3); //--- get correct timestamp offset
           //timestampOffset := findPos1 - 166*3;
           if timeStampOffset <> -1 then  //--- get the timestamp
           begin
               timestamp := copy(l, timestampOffset + 5*3, 3) +
                            copy(l, timestampOffset + 4*3, 3) +
                            copy(l, timestampOffset + 3*3, 3) +
                            copy(l, timestampOffset + 2*3, 3);
               try
                 timestamp := datetimetostr((strToInt(RemoveSpaces('$' + timestamp)) / 86400) + UnixStartDate);
               except
                 timestamp := '';
               end;
           end else timestamp := '';


           //--- extract the message
           l := copy(l, findPos1 + length(tokenConvStart), length(l)); //--- cut the string starting from the beginning of the message

           findPos2 := pos(tokenConvEnd, l); //--- now look where conversaion ends (null byte $00)
           if findPos2 > 0 then  //--- found!
           begin
             messg := trim(ConvertASCIIHexToString(copy(l, 1, findPos2)));
             if messg <> '' then
             begin
               if copy(l, findPos2 + 3, 3*2) = '00 39 ' then findPos2 := findPos2 + 3*10; //-- offset correction when we are dealing with files

               conversation := timestamp + ' [' + MatchParticipantsById(p, currentUserId, ownName) + ']: ' + messg;
               if conversationIsValid(conversation) then myOutputList.Add(conversation);
             end;
             
           end;
         end;
      until findPos1 = 0;
  end;
end;






//----------------------------------get all skype conversations from sorted list
function retrieveConversations(skypeFiles: TSkypeChatSyncFiles; ownID: string): TSkypeConversations;
var i, convID: integer;
    s: TStringList;
    part: TParticipants;
    MySkypeConversations: TSkypeConversations;

  function getConversationID(participants: TParticipants): integer;
  var i: integer;
  begin
    result := -1;
    for I := 0 to length(MySkypeConversations) - 1 do
    begin
      if (((participants.people[0] = MySkypeConversations[i].participants.people[0]) and (participants.people[1] = MySkypeConversations[i].participants.people[1])))
         or
         (((participants.people[0] = MySkypeConversations[i].participants.people[1]) and (participants.people[1] = MySkypeConversations[i].participants.people[0]))) then
         begin
           result := i;
           break;
         end;
    end;
  end;
  
begin
  setLength(MySkypeConversations, 0);

  //--- make list of all possible conversations
  for i := 0 to length(skypeFiles) - 1 do
  begin
    part := getConvParticipants(skypeFiles[i].fileName);
    convID := getConversationID(part);
    if convID = -1 then
    begin
      setLength(MySkypeConversations, length(MySkypeConversations) + 1);
      MySkypeConversations[length(MySkypeConversations) - 1].participants := part;
      MySkypeConversations[length(MySkypeConversations) - 1].content := TStringList.Create;
    end;
  end;

  //--- build conversation list
  for i := 0 to length(skypeFiles) - 1 do
  begin
    part := getConvParticipants(skypeFiles[i].fileName);
    convID := getConversationID(part);
    MySkypeConversations[convID].content.Add('-------------------');
    //MySkypeConversations[convID].content.Add(part.id[0]); //--- needed for tests only
    MySkypeConversations[convID].content.Add(datetimetostr(skypeFiles[i].fileDateTime) + ' (' + extractFileName(skypeFiles[i].fileName) + ')');
    MySkypeConversations[convID].content.Add('-------------------');

    s := TStringList.Create;
    getConversation(skypeFiles[i].fileName, s, part, ownID);
    MySkypeConversations[convID].content.Add(s.Text);
    s.Free;

  end;

  result := MySkypeConversations;
end;







end.
